module wei_datum_base
  use wei_basics_mod
  use wei_slow_mod
  use wei_imaging_mod
  use wei_data_mod
  use wei_wavespace_mod
  use wei_fft_mod
  use wei_wc_mod
  use wei_kmap_mod
  use wei_xmap_mod
  use wei_datum_z_mod
  implicit none
  type(sep3d), private, save :: wsep,ssep
  complex, pointer :: wfld(:,:,:,:,:)
  logical, private, save :: cd,cr


contains
!!$
!!$=head1 
!!$
!!$ weidatum_init - Initialize datuming operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$  call weidatum_init(SLin,WCin,FKin,FXin,IGin,in6,in7,in8,in9)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item SLin - Slowness operator init  L<weislo1_init>
!!$
!!$=item WCin - WCop operator init  L<weikxmig1_init>
!!$
!!$=item FKin - FK operator init  L<weiwem_init>
!!$
!!$=item FXin - FX operator init  L<weissf_init>
!!$
!!$=item in6,..in9 -  integer,optional
!!$
!!$ Dummy operators for future expansion
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize datuming operator
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weimig> L<wei>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  subroutine weidatumz_init(eraseDOWN,eraseD,SLin,WCin,FKin,FXin,in6,in7,in8,in9)
    integer,external,optional ::                     in7,in8,in9,in6
    integer :: st,er,ierr
    type(sep3d) :: dsep
    type(sep3d) :: down_sep
    character(len=128) :: stag
    logical  :: eraseDOWN,eraseD
#ifdef DEBUG
    character(len=128), parameter :: unit='weidatum_init'
#endif
    interface !!-------------------------
       integer function    SLin() result(st)
       end function SLin
       integer function WCin() result(st)
       end function WCin
       integer function    FKin() result(st)
       end function FKin
       integer function    FXin() result(st)
       end function FXin
    end interface !!---------------------

    if(.not. init_wei_basics()) &
      call seperr("trouble initializing wei basics")

    !INITIALZE WAVEFIELD SPACE
    if(my_verb()) write(0,*) "BEFORE INITIALIZE WAVE SPACE"
    if(.not. init_reg_wave_space(wsep)) &
      call seperr("trouble initializing wave space")
    call store_current_wave_space(wsep)

    !INITIALIZE DATUMED DATA
    if(my_verb()) write(0,*) "BEFORE INITIALIZE DATUMED DATA"
    if(.not. init_data(erasedown,"DOWN",down_sep,CR)) &
      call seperr("trouble initializing down data")

    !INITIALIZE DATA
    if(my_verb()) write(0,*) "BEFORE INITIALIZE DATA"
    if(.not. init_data(erased,"D",dsep,CD)) &
      call seperr("")

      

    !INITIALZE SLOWNESS
    if(my_verb()) write(0,*) "BEFORE INITIALIZE SLOWNESS"
    if(.not. init_slow()) &
      call seperr("")

    !INITIALZE SLOWNESS FROM FILE
    if(my_verb()) write(0,*) "BEFORE INITIALIZE SLOWNESS FILE"
    if(.not. init_slow_tag("S",stag,ssep)) &
      call seperr("")

    call store_slowness_masters((/ssep/),(/stag/))

    !INITIALIZE SLOWNESS OPERATOR
    if(my_verb()) write(0,*) "BEFORE INITIALIZE SLOWNESS OPERATOR"
    st=SLin()     !! slownesses init


    !INITIALIZE KMAPS
    if(my_verb()) write(0,*) "BEFORE INITIALIZE MAPS"
    call weikmap_init(wsep)
    call weixmap_init(wsep)
    call show_mem("map")

    if(my_verb()) write(0,*) "BEFORE INITIALIZE FK"
    st=FKin()     !! FK op init
    if(my_verb()) write(0,*) "BEFORE INITIALIZE FX"
    st=FXin()     !! FX op init
    if(my_verb()) write(0,*) "BEFORE INITIALIZE WC"
    st=WCin()     !! WCop init
    if(my_verb()) write(0,*) "BEFORE INITIALIZE PATH"
    st=Path_z_init()   !! Path init

    if(my_verb()) write(0,*) "BEFORE SPACE LOGIC "
    if(.not. check_space_logic(wsep,(/dsep,down_sep/),ssep=(/ssep/))) &
    	call seperr("spaces don't conform" )
     
    if(my_verb()) write(0,*) "BEFORE MEMORY ALLOCATION"
    if(.not. wem_allocate_memory(wsep,my_nomp()))&
      call seperr("trouble allocating memory")

    !INITIALIZE FFTS after memory allocation
    if(my_verb()) write(0,*) "BEFORE FFT INIT"
    call assign_wave_space(wfld)  !get memory for wfld
    call weifft_init(wfld)  
    call clear_wave_space()  !get memory for wfl
    if(my_verb()) write(0,*) "BEFORE STRUCTURE DELETE"
    call sep3df_delete(dsep,"D")
    call sep3df_delete(down_sep,"DOWN")

#ifdef DEBUG
    if(my_debug()) call out(unit)
#endif
  end subroutine

  integer function weidatumz(adj,add,down_,dat_,SLop,  WCop,FKop,FXop,op6,op7,op8,op9)
  logical,            intent(in) :: adj,add
  character(len=128), intent(in) :: down_,dat_
  integer,external,optional :: WCop,FKop,FXop,op6,op7,op8,op9
  real , pointer :: dstep(:),omega(:)
  integer :: i,path,ierr,iz,iw,ith,iws,izs
  character(len=60) :: string,wave_out
  type(sep3d) :: dsep,down_sep
  logical :: ipath,logic
  type(sep3d) :: r
  integer,external :: srite
  integer :: nwind(5),fwind(5)
#ifdef DEBUG
    character(len=128), parameter :: unit='weidatums'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
                                                                                           
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif
  
   call init_sep3d(dat_,dsep,"SCRATCH")
   call init_sep3d(down_,down_sep,"SCRATCH")
   call store_current_data_space(dsep)
   call store_current_data_tag(dat_)
   call store_current_add(add)
   call store_current_adj(adj)


!   if(.not. operation_check(adj,cD,cR)) &
!     call seperr("")

   call store_current_down_space(down_sep)
   call store_current_down_tag(down_)
    
   if(.not. my_restart()) then
     if(.not. add)  then
       if(.not. adj) then
          if(.not. zero_data(dat_,dsep)) &
           call seperr("trouble zeroing data")
       else
         if(.not. zero_data(down_,down_sep)) &
           call seperr("trouble zeroing datumed data")
       end if
     end if
   end if
   call sep_begin_prog() !it is now safe for other jobs to be started on 
      !this thread


   call assign_wave_space(wfld)  !get memory for wfld


   nullify(dstep,omega)
   if(.not. path_z_begin()) call seperr("trouble initializing path")


   string="datuming down"
   if(adj) string="datuming up"


     if(.not. path_z_op(mode_begin,wfld,dstep,omega)) &!Grab the first block
       call seperr("trouble grabbing first path block")

     do while(associated(omega))  !{Go until we hit an error or are finished
        call store_current_s_r(.true.)
        ierr=SLop()
        do iw=1,size(omega) !{ loop over the frequencies in the block
          call store_current_omega(omega(iw))
          ith=1
          do iz=1,size(dstep) !{ loop over the the depths in this block
            call store_current_izs(iz)
            call store_current_dstep(dstep(iz))
            call  weikxmig_1wave(wfld,iw,iz,ith,WCop,FKop,FXop)
            call print_progress(string)
          end do ! }
        end do! }
        if(.not.path_z_op(mode_end,wfld,dstep,omega)) &
            call seperr("Trouble finishing block")

       if(.not.path_z_op(mode_begin,wfld,dstep,omega)) then
          call seperr("Trouble grabbing next block")
       end if
     end do !}


   call clear_wave_space()  !get memory for wfld

    
!if(sep_thread_num()==0) call print_timers()
    weidatumz=0
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function
end module 
